import dnf
import os
import re
import io
import logging
import asyncio
import concurrent
from typing import List, Dict

from .log import red_log


async def to_process(func, *args, **kwargs):
    with concurrent.futures.ProcessPoolExecutor() as pool:
        return await asyncio.get_event_loop().run_in_executor(pool, func, *args, **kwargs)

class dnfBaseConf(dnf.Base):
    "pickable wrapper on dnf.Base"

    def __init__(self, cachedir, arch, installroot, urls):
        self.cachedir = cachedir
        self.urls = urls
        basearch = arch
        self.id = "{}-{}".format(arch, re.sub(r'[:/.$]', '_', urls[0]["url"]))

        conf = dnf.conf.Conf()
        conf.cachedir = os.path.join(cachedir, "dnf", self.id)
        conf.arch = arch
        conf.basearch = basearch
        conf.installroot = installroot

        super().__init__(conf=conf)

        id_i = 0
        for urlpriority in self.urls:
            self.repos.add_new_repo(
                "{}-{}".format(self.id, id_i), self.conf, baseurl=[urlpriority["url"]], module_hotfixes=True, priority=urlpriority["priority"])
            id_i += 1

# fill sack arch, independant to be execute in separate process
def _fill_sack_arch(cachedir, arch, installroot, urls):
    base = dnfBaseConf(cachedir, arch, installroot, urls)
    ret = True
    logger = logging.getLogger('dnf')
    logger.propagate = False
    try:
        # check dnf output
        with io.StringIO() as buf:
            handler = logging.StreamHandler(buf)
            logger.addHandler(handler)
            base.update_cache()
            base.fill_sack(load_system_repo=False)
            output = buf.getvalue()
        # looking for module platform in dnf output
        # and add it into conf
        regex = re.search(
            r"nothing provides module\((\w+:\w+)\)", output)
        if regex:
            base.conf.module_platform_id = regex.group(1)
            base.update_cache()
            base.fill_sack(load_system_repo=False)
    except dnf.exceptions.RepoError as e:
        red_log.warning(
            "repo %s arch %s failed to sack! (Exception:%s)", urls, arch, e)
        ret = False
    finally:
        logger.propagate = True
    return ret


class Dnf:
    def __init__(self, urls, arches, cachedir):
        self._urls = urls
        self._arches = arches
        self._cachedir = cachedir
        os.makedirs(self._cachedir, exist_ok=True)
        self._base_arches = self._get_base_arches()

    @property
    def urls(self):
        return self._urls

    def _get_base_arches(self) -> Dict[str, dnfBaseConf]:
        base_arches: Dict[str, dnfBaseConf] = {}
        for arch in self._arches:
            try:
                base_arches[arch] = dnfBaseConf(self._cachedir, arch, "/tmp", self._urls)
            except dnf.exceptions.Error as e:
                red_log.error("dnf error on arch %s, error=%s", arch, e)
        return base_arches

    async def fill_sack(self, base, arch):
        # fill sack into a separate process (preventing from blocking event loop)
        if not await to_process(_fill_sack_arch, base.cachedir, arch, base.conf.installroot, base.urls):
            return
        # update sack in our base object (real sack was done in separate process)
        await asyncio.to_thread(base.fill_sack, load_system_repo=False)

    async def fill_sacks(self):
        if not self._base_arches:
            return False
        red_log.debug("fill sack %s", self._urls)
        await asyncio.gather(*[asyncio.create_task(self.fill_sack(base, arch)) for arch, base in self._base_arches.items()])

    async def revision(self):
        revision = 0
        for _, base in self._base_arches.items():
            for r in base.repos.iter_enabled():
                r.load()
                r = r._repo.getRevision() or 0
                revision += int(r)
        return revision

    async def query(self, arch):
        base = self._base_arches.get(arch, None)
        if not base:
            return None

        if not base.sack:
            await self.fill_sacks()

        return await asyncio.to_thread(base.sack.query)


    async def filter(self, **kwargs):
        tasks = []
        for arch, base in self._base_arches.items():
            if not base:
                red_log.warning("No base dnf for arch=%s urls=%s", arch, self.urls)
                continue

            # if one miss fill_sacks all
            if not base.sack:
                await self.fill_sacks()

            def _filter(base, **kwargs):
                return base.sack.query().filter(**kwargs)

            tasks.append(asyncio.create_task(asyncio.to_thread(_filter, base, **kwargs)))


        await asyncio.gather(*tasks, return_exceptions=True)
        res = []
        for task in tasks:
            res += list(task.result())
        return res
