from unittest import result
from .socketio_wrapper import socketio
import asyncio
import secrets
import inspect
from datetime import datetime, timedelta

from cookies_events_manager import cookies_events_manager as cem
from cookies_events_manager import WSCookieEventsManager

from .error import (RedpeskUtilsError, formaterror)
from .log import red_log

TICKET_SIZE = 16


class EventsError(RedpeskUtilsError):
    pass


class Ticket:
    def __init__(self, events, event_name, request, ticketid=None, task: asyncio.Task = None):
        self._events = events
        self._event_name = event_name
        self._ticketid = ticketid if ticketid else secrets.token_urlsafe(TICKET_SIZE)
        self._request = request
        self._task = task
        if self._task:
            self._task.add_done_callback(self.done_cb)
        self._done = False
        self._error = None
        self._res = {}

    def done_cb(self, task):
        result = None
        error = None
        try:
            result = task.result()
        except Exception:
            error = formaterror()

        self._events.new_event_cb(self._event_name, self._ticketid, result, True, error)

    @property
    def ticketid(self):
        return self._ticketid

    @property
    def done(self) -> bool:
        return self._done

    @done.setter
    def done(self, done: bool) -> None:
        self._done = done

    @property
    def error(self) -> str:
        return self._error

    @error.setter
    def error(self, error: str) -> None:
        self._error = error

    @property
    def res(self) -> dict:
        return self._res

    @res.setter
    def res(self, res: dict) -> None:
        self._res = res

    def format(self) -> dict:
        return {
            "ticket": self._ticketid,
            "done": self._done,
            "error": self._error,
            "res": self._res,
            "request": str(self._request) if self._request else None,
            "args": str(self._request.get("datajson", "") if self._request else None)
        }


class RedEvents:
    def __init__(self, cookie_name):
        self._cookie_name = cookie_name
        # async server
        self._sio = socketio.AsyncServer(
            async_mode='aiohttp', cookie=self._cookie_name)
        self._work_queue = asyncio.Queue()
        self._cookie_manager = WSCookieEventsManager(self._work_queue)

        # tickets: each event is matching a ticket
        self._tickets = {}

        self._bck_task = self.start_background_task(self._background_task)

        self._sio.on('connect', handler=self.connect)
        self._sio.on('disconnect', handler=self.disconnect)

    def stop_bck_task(self):
        self._bck_task.cancel()
        return self._bck_task

    ### HANDLE TICKETS BY EVENT ###
    # get ticket by event
    def get_event_tickets(self, event_name, strict=False):
        event_tickets = self._tickets.get(event_name, None)
        if not event_tickets and strict:
            raise RedpeskUtilsError("No ticket for event name {}".format(event_name))
        return event_tickets

    def _add_ticket_event(self, request, event_name, ticketid=None, task=None):
        ticket_events = self.get_event_tickets(event_name)
        if not ticket_events:
            ticket_events = {}
            self._tickets[event_name] = ticket_events
        ticket = Ticket(self, event_name, request, ticketid=ticketid, task=task)
        ticket_events[ticket.ticketid] = ticket
        return ticket, ticket_events

    def add_ticket_event(self, request, event_name, ticketid=None, task=None):
        ticket, _ = self._add_ticket_event(request, event_name, ticketid=ticketid, task=task)
        return ticket

    def _get_ticket_from_event_tickets(self, event_tickets, ticketid, strict=False) -> Ticket:
        ticket = event_tickets.get(ticketid, None)
        if not ticket and strict:
            raise RedpeskUtilsError("No ticket for ticketid {} (event name {})".format(ticketid, event_name))
        return ticket

    def _get_ticket(self, event_name, ticketid, strict=False):
        event_tickets = self.get_event_tickets(event_name, strict=strict)
        if not event_tickets:
            return None, None
        return self._get_ticket_from_event_tickets(event_tickets, ticketid, strict=strict), event_tickets

    def get_ticket(self, event_name, ticketid, strict=False) -> Ticket:
        ticket, _ = self._get_ticket(event_name, ticketid, strict=strict)
        return ticket

    def attach(self, app):
        self._sio.attach(app)

    # sending event ticket
    #   create: allow to create a ticket if no ticket is matching
    def new_event_cb(self, event_name, ticketid, data, done, error, create=False):
        ticket, event_tickets = self._get_ticket(event_name, ticketid, strict=not create)
        red_log.debug("[NEW EVENT CB %s]: ticketid:%s done:%s error:%s \n\tDATA=%s\n\tTICKET=%s",
                      event_name, ticketid, done, error, data, ticket.format() if ticket else None)

        if not ticket:
            if not create:
                raise RedpeskUtilsError("No ticket for ticketid {} (event name {})".format(ticketid, event_name))
            ticket, event_tickets = self._add_ticket_event(None, event_name, ticketid=ticketid)

        if data and isinstance(data, dict) and isinstance(ticket.res, dict):
            ticket.res.update(data)
        else:
            ticket.res = data
        ticket.done = done
        ticket.error = error
        red_log.debug("[SENDING EVENT %s]: %s", event_name, ticket.format())
        self._cookie_manager.new_event(event_name, ticket.format())
        # remove ticket if done: (TODO: maybe keep it in cache if needed)
        if done:
            event_tickets.pop(ticketid)

    def start_background_task(self, target, *args, **kwargs):
        return self._sio.start_background_task(target, *args, **kwargs)

    # init events
    def connect(self, sid, environ):
        http_cookie = environ.get("HTTP_COOKIE", None)
        if not http_cookie:
            raise EventsError(
                "connect websocket: No entry for sid={} HTTP_COOKIE in environ={}".format(sid, environ))
        if(self._cookie_name in http_cookie):
            # LOG utility -----------------------------------------------------
            red_log.debug("[WS Connect] This client already got a cookie: " +
                          environ["HTTP_COOKIE"])
            # -----------------------------------------------------------------

            # Get the cookie
            cookie_index = (environ["HTTP_COOKIE"].find(self._cookie_name) +
                            len(self._cookie_name) + 1)
            sub_cookie = environ["HTTP_COOKIE"][cookie_index:]
            client_cookie = sub_cookie.split(sep=';')[0]

            # Update the client SID
            self._cookie_manager.modify_client_sid(client_cookie, sid)

            # LOG utility -----------------------------------------------------
            red_log.info("[WS Connect] WebSocket connect accepted from cookie: " +
                         str(client_cookie))
            # -----------------------------------------------------------------

        else:
            raise ConnectionRefusedError("[WS Connect]" +
                                         " You need to have a valid cookie")

    def disconnect(self, sid):
        if self._cookie_manager.delete_client_from_sid(sid):
            red_log.info("disconnect: Remove client %s", sid)
        else:
            red_log.warning("disconnect: Cannot remove client %s", sid)

    def add_client_if_needed(self, cookie):
        if self._cookie_manager.is_cookie_valid(cookie):
            return cookie

        cookie = secrets.token_urlsafe(64)
        expiration_date = (datetime.utcnow() + timedelta(days=15000))
        self._cookie_manager.add_client(cookie, None, None, expiration_date)
        return cookie

    def register(self, cookie, eventName):
        self._cookie_manager.subscribe(cookie, eventName)

    def unregister(self, cookie, eventName):
        self._cookie_manager.unsubscribe(cookie, eventName)

    async def _background_task(self):
        """Background task that runs in parallel of the API execution.
        It handles the following tasks : disconnecting expired clients \
        and sending events to given SID.

        :param work_queue: A Queue (FIFO) storing the tasks to execute
        :type work_queue: asyncio.Queue
        """
        try:
            cond_back_task = True
            while cond_back_task:
                work_to_do = await self._work_queue.get()
                # The task is to disconnect a client
                if(work_to_do[cem.TASK_KEY] == cem.TASK_DISCONNECT_CLIENT):
                    sid_to_disconnect = work_to_do[cem.TASK_DATA_KEY]

                    # LOG utility ---------------------------------------------
                    red_log.info("[BackGnd Task] Disconnecting SID : %s",
                                 sid_to_disconnect)
                    # -----------------------------------------------------------------

                    # Disconnect the corresponding SID
                    await self._sio.disconnect(sid_to_disconnect)

                # The task is to emit an event to a client
                elif(work_to_do[cem.TASK_KEY] == cem.TASK_EMIT_EVENT):
                    # Get the SID on which the event will be send
                    sid_to_use = work_to_do[cem.TASK_DATA_KEY][cem.TASK_EVENT_SID_KEY]
                    # Get the event identifier
                    event_id = work_to_do[cem.TASK_DATA_KEY][cem.TASK_EVENT_ID_KEY]
                    # Get the data to send
                    data_to_send = work_to_do[cem.TASK_DATA_KEY][cem.TASK_EVENT_DATA_KEY]

                    # LOG utility ---------------------------------------------
                    data_to_send_str = str(data_to_send)
                    red_log.debug(
                        "[BackGnd Task] Sending the event %s to SID %s\n\t\t-> %s ... %s",
                        event_id, sid_to_use, data_to_send_str[:50], data_to_send_str[-50:])
                    # -----------------------------------------------------------------

                    # Send the event
                    await self._sio.emit(event_id, data_to_send, room=sid_to_use)

                # The task is to end the background task
                elif (work_to_do[cem.TASK_KEY] == cem.TASK_END):
                    cond_back_task = False

                # Notify the queue that the "work item" has been proceed
                self._work_queue.task_done()
        except asyncio.CancelledError:
            await self._stop_background_task()

        red_log.debug("[Event BackGnd Task] Task ended")

    async def _stop_background_task(self):
        for value in self._cookie_manager.subscriptions.values():
            await self._sio.close_room(value[cem.SID_KEY])
        await self._sio.eio.disconnect(True)

        # Stop the module cookie_events_manager.py
        self._cookie_manager.end_manager()

        # Need to cancel this asyncserver tasks
        for task in asyncio.all_tasks():
            if task._coro.__qualname__ not in (
                    "AsyncServer._service_task",
                    "AsyncSocket.schedule_ping.<locals>.send_ping",
                    "WebSocket.wait"
            ):
                continue
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass


def ticket(fgetevents, event_name):
    """ decorator for handle async request
           - request should return: res, task
           - events is the RedEvents class, below a way to give it
                @ticket(lambda self: self._events, "MYEVENT")
    """
    def dec(func):
        async def inner(self, request, *args, **kwargs):
            events: RedEvents = fgetevents(self)
            try:
                task, res = await func(self, request, *args, **kwargs)
            except Exception:
                raise
            ticket = events.add_ticket_event(request, event_name, task=task)
            ticket.res = res
            events.new_event_cb(event_name, ticket.ticketid, ticket.res, False, None)
            return ticket.format()
        inner.__name__ = func.__name__
        inner.__signature__ = inspect.signature(func)
        return inner
    return dec


def ticket_setid(fgetevents, event_name):
    """ decorator for handle async request by setting ticketid
           - request should return: task, res, ticketid (task can be None)
           - events is the RedEvents class, below a way to give it
                @ticket(lambda self: self._events, "MYEVENT")
    """
    def dec(func):
        async def inner(self, request, *args, **kwargs):
            events: RedEvents = fgetevents(self)
            try:
                task, res, ticketid = await func(self, request, *args, **kwargs)
            except Exception:
                raise
            ticket = events.add_ticket_event(request, event_name, ticketid=str(ticketid))
            ticket.res = res
            return ticket.format()
        inner.__name__ = func.__name__
        inner.__signature__ = inspect.signature(func)
        return inner
    return dec
