import os
import re
import time
import logging
import shutil
import git
import datetime
import contextlib
import tempfile
from enum import Enum
import concurrent.futures

from . import (Spec, ENTRY_HEXSHA, from_string)
from .error import (ArchiveAlreadyThereError, NoArchiveNameVersionError, GitSourceError)

from pyrpm.spec import replace_macros


class RefType(Enum):
    SHA = 1
    BRANCH = 2
    TAG = 3


class GitSource:
    def __init__(self, workdir, log=None):
        self._pattern_version_tag = r"^v{0,1}(\d+(\.\d+){0,2})$"
        self.REGEX_DESCRIBE = r"(.*?)(-(\d+)-g(\w+))?"
        self.REGEX_VERSION_SPECFILE = r"\d+(\.\d+){2}"
        self.REGEX_GITREV = r"g([a-f0-9]{7})"
        self.REGEX_GITREV_CHECKOUT = r"[a-f0-9]{7,40}"
        self.REGEX_ARCHIVE_EXTENSION = r"^(.*)-(\d+\.\d+.\d+(\+\d{8}(\+\d{6})?\+\d+\+g\w{7})*)(\..+)$"
        self._workdir = workdir
        self._env_git = {"GIT_TERMINAL_PROMPT": "0"}
        if log:
            self.log = log
        else:
            self.log = logging.getLogger("GitSource")

        self._gitsourcedir = os.path.join(
            os.path.join(self._workdir, "gitsource"))
        self._copytags = None

        self._spec = None
        self._path_source = None
        self.repo_gitsource = None
        self._default_branch = None
        self._commit_checkout = None
        self._ref_type = RefType.BRANCH
        self._archive_pattern = "%{name}-%{version}"
        self._archive_extension = None

    @property
    def current_commit(self):
        return self.repo_gitsource.head.commit

    #############
    # REGEX
    #############
    def _is_regex_gitrev_checkout(self):
        return True if re.fullmatch(
            self.REGEX_GITREV_CHECKOUT, self._commit_checkout) else False

    def _regex_describe(self, describe):
        tag_re = re.fullmatch(self.REGEX_DESCRIBE, describe)
        if not tag_re:
            return (None, None, None)

        tagv = tag_re.group(1)
        nb_commit = tag_re.group(3)
        nb_commit = int(nb_commit) if nb_commit else 0
        gcommit = tag_re.group(4)

        tagv_re = re.fullmatch(self._pattern_version_tag, tagv)

        if tagv_re:
            tagv = self.adjust_tagv(tagv_re.group(1))
        else:
            tagv = None

        return (tagv, nb_commit, gcommit)

    def _regex_gitrev_specfile(self, version):
        name_re = re.search(self.REGEX_GITREV, version)
        if not name_re:
            return None

        return name_re.group(1)

    def _is_regex_version_specfile(self, version):
        return True if re.fullmatch(
            self.REGEX_VERSION_SPECFILE, version) else False

    # set commit checkout:
    #       load spec object if exists
    #       if no commit try to get commit from specfile
    def _set_commit_checkout(self, commit):
        self._commit_checkout = commit if commit else self._default_branch
        self.log.info(
            "Will try to checkout to %s (commit_checkout)", self._commit_checkout)

    # set spec structure
    #   if a spec_content is provided, fill the _spec attribute
    def _set_spec(self, spec_content):
        if spec_content:
            self._spec = from_string(spec_content)

    # set path source:
    #       if no path source get the one from specfile
    def _set_path_source(self, path_source):
        if not path_source:
            if not self._spec or not self._spec.url:
                raise GitSourceError("Sources cannot be found!! \n"
                                     "Indeed, there is no url set in your "
                                     "specfile(URL: ) and "
                                     "no path source is passed in argument")
            self._path_source = self._spec.url
        else:
            self._path_source = path_source

    def _is_shallow_clone(self):
        return os.path.exists(self._gitsourcedir + "/.git/shallow")

    # shallow clone: depth=1, single-branch
    def _shallow_clone(self):
        if self._type_ref == RefType.SHA:
            ref=self._default_branch
        else:
            ref=self._commit_checkout
        try:
            self.log.info("Trying shallow clone source from %s:%s into %s",
                          self._path_source, self._commit_checkout,
                          self._gitsourcedir)
            self.repo_gitsource = git.Repo.clone_from(
                self._path_source, self._gitsourcedir, branch=ref,
                env=self._env_git,
                depth=1, single_branch=True)
            self.log.info("Successfull shallow cloned.")
        except git.exc.GitCommandError as e:
            if "does not support shallow" in str(e):
                self.repo_gitsource = git.Repo.clone_from(
                    self._path_source, self._gitsourcedir, branch=ref,
                    env=self._env_git,
                    no_single_branch=True)
            else:
                os.rmdir(self._gitsourcedir)
                if e.args[1] == 128:
                    raise GitSourceError(
                        "Invalid git reference: {}".format(self._commit_checkout))
                else:
                    raise GitSourceError(
                        "Failed to clone {}. Not a valid git repository.".format(self._path_source))
        if self._type_ref == RefType.SHA:
            if not self._go_checkout():
                raise GitSourceError(
                    "Invalid git reference: {}".format(self._commit_checkout))

    # set gitsource localdir: if path source is localdir: set gitsourcedir to it
    #       forbidden to clean in this case
    def _set_gitsource_localdir(self, clean):
        self._gitsourcedir = self._path_source
        try:
            self.repo_gitsource = git.Repo(self._gitsourcedir)
        except (git.exc.GitCommandError, git.exc.InvalidGitRepositoryError):
            raise GitSourceError(
                "%s is not a valid git repository" % self._gitsourcedir)

        if clean:
            raise GitSourceError(
                "Clean forbidden when --path_source is a local directory, "
                "aborting...")

    # set gitsource: try to set repo from gitsourcedir
    #       if failed: clean it
    #       check also that the remote is still valid: else clean
    def _set_gitsource(self, clean):
        try:
            self.repo_gitsource = git.Repo(self._gitsourcedir)
        except (git.exc.GitCommandError, git.exc.InvalidGitRepositoryError):
            clean = True

        # check remote is present in repo
        if not clean:  # if clean, gitsource will be removed anyway
            clean = True

            # check url from remote origin
            for url in self._get_remote_origin().urls:
                if self._path_source.strip() == url.strip():
                    clean = False

            # if still clean: warning
            if clean:
                self.log.warning(
                    "Remote: --path_source={} not found in remote in repo {},"
                    "gitsource will be cleaned".format(self._path_source, self._gitsourcedir))
                self.repo_gitsource = None
        return clean

    # checkout success: return True or False
    def checkout_success(self):
        try:
            self.repo_gitsource.git.checkout(self._commit_checkout)
            self.log.info("Successful checkout to %s", self._commit_checkout)
        except git.exc.GitCommandError:
            return False
        return True

    def _set_default_branch(self):
        re_branch = r"refs/heads/(\S*)"
        branch = re.search(re_branch, git.cmd.Git().ls_remote("--symref",
                                                              self._path_source, "HEAD", env=self._env_git))
        if branch:
            self._default_branch = branch.group(1)
        else:
            self._default_branch = "main"

    def _get_depth(self):
        return self.repo_gitsource.head.commit.count(first_parent=True)

    def _count_all(self):
        return self.repo_gitsource.head.commit.count()

    # git fetch depth: fetch increasing depth (shallow cloned)
    def _git_fetch_depth(self, depth):
        ref=self._commit_checkout
        if self._type_ref == RefType.TAG:
            type="tags"
        else:
            type="heads"
            if self._type_ref == RefType.SHA:
                ref=self._default_branch
        self.log.info("Fetching %s with depth %d", ref, depth)
        self._get_remote_origin().fetch(f"refs/{type}/{ref}:refs/{type}/{ref}", depth=depth, quiet=True, prune=True, prune_tags=True, update_head_ok=True)

    # git fetch depth callback: call callback after each fetch
    def _git_fetch_depth_callback(self, maxdepth, callback, *args):
        # first trying to fetch deeper and deeper: avoiding removing shallow
        # clone
        while self._count_all() < maxdepth:
            previous = self._count_all()
            self._git_fetch_depth(self._count_all()+10)
            if callback(*args):
                return True
            if previous >= self._count_all():
                break
        return False

    # go checkout: special checkout for shallow clone
    #       go deeper and deeper
    def _go_checkout(self):
        if self.checkout_success():
            return True

        if not self._is_shallow_clone():
            return False

        if self._type_ref is not RefType.SHA:
            return False

        # first trying to fetch deeper and deeper: avoiding removing shallow
        # clone
        if self._git_fetch_depth_callback(200, self.checkout_success):
            return True

        # final try: unshallow
        try:
            self.log.info(
                "Unshallowing repo because SHA %s not found within the last %d commit of %s", self._commit_checkout, 200, self._default_branch)
            if self._is_shallow_clone():
                self._get_remote_origin().fetch(unshallow=True)
            self.repo_gitsource.config_writer().set_value("remote \"origin\"", "fetch", "+refs/heads/*:refs/remotes/origin/*").release()
            self._get_remote_origin().fetch()

        except git.GitCommandError:
            pass

        return self.checkout_success()

    # get remote origin: only work with origin remote
    def _get_remote_origin(self):
        return self.repo_gitsource.remotes.origin

    # check if commit checkout is available in ls remote
    def _git_check_ls_remote(self, options):
        return True if git.cmd.Git().ls_remote(options, self._path_source,
                                               self._commit_checkout,
                                               env=self._env_git) else False

    # check if commit checkout is available in ls remote heads
    def _git_check_ls_remote_heads(self):
        return self._git_check_ls_remote("--heads")

    # check if commit checkout is available in ls remote tags
    def _git_check_ls_remote_tags(self):
        return self._git_check_ls_remote("--tags")

    # check if at least one tag is matching regex with git ls-remote
    def _git_check_ls_remote_match_tag_regex(self):
        re_tag = r"refs\/tags\/(.*)"
        for line in git.cmd.Git().ls_remote("--tags", "--refs",
                                            self._path_source, env=self._env_git).splitlines():
            tag = re.search(re_tag, line)
            if tag:
                tag_ok = re.search(self._pattern_version_tag, tag.group(1))
                if tag_ok:
                    return True
        return False

    # is branch: check if commit_checkout is a branch
    #       2 modes: from ls remote or from local repo
    def _is_branch(self):
        upstream = True
        if self.repo_gitsource:
            if self._path_source == self._workdir:
                if self._commit_checkout in [branch.name for branch in self.repo_gitsource.branches]:
                    upstream = False
        if upstream:
            if not self._git_check_ls_remote_heads():
                return False
        self._type_ref = RefType.BRANCH
        return True

    # is tag: check if commit_checkout is a tag
    #       2 modes: from ls remote or from local repo
    def _is_tag(self):
        upstream = True
        if self.repo_gitsource:
            if self._path_source == self._workdir:
                if self._commit_checkout not in [str(tag)
                                        for tag in self.repo_gitsource.tags]:
                    upstream = False
        if upstream:
            if not self._git_check_ls_remote_tags():
                return False
        self._type_ref = RefType.TAG
        return True

    # set ref type: set the type of commit checkout
    def _set_ref_type(self):
        # is branch ?
        is_set = self._is_branch()

        # is tag ?
        if not is_set:
            is_set = self._is_tag()

        # is looks like a sha ?
        if not is_set:
            is_set = self._is_regex_gitrev_checkout()
            self._type_ref = RefType.SHA

        if not is_set:
            raise GitSourceError(
                "Invalid git reference {} for {}".format(
                    self._commit_checkout, self._path_source))

        self.log.info(
            "Commit checkout %s is set as a %s ref", self._commit_checkout,
            self._type_ref)

    def _need_redo_archive(self, force, old_version, old_extension):
        if force:
            return

        # all is indexed ?
        if self.repo_gitsource.index.diff("HEAD"):
            self.log.debug("Uncommited changes in %s", self._gitsourcedir)
            self._version_tag = "%s+dirty" % self._version_tag
            return

        if self._spec.sha == self.repo_gitsource.head.commit.hexsha and self._archive_extension == old_extension:
            if self._version_tag and self._version_tag != old_version:
                return
            raise ArchiveAlreadyThereError(
                "Commit in specfile {} and current commit after gitsource {} are the same, no need to pkg archive".format(
                    self._spec.sha, self.repo_gitsource.head.commit.hexsha))

    def _set_specfile(self, spec_content, previous_archive_name, force):
        specfile = from_string(spec_content)

        self._check_archive_specfile(specfile)

        self._spec.sha = specfile.sha

        if not previous_archive_name:
            return specfile

        previous_ext_re = re.search(
            self.REGEX_ARCHIVE_EXTENSION, previous_archive_name)

        if not previous_ext_re:
            return specfile

        # check old_name :! specfile name
        if previous_ext_re.group(1) != specfile.name:
            return specfile
        old_version = previous_ext_re.group(2)
        old_extension = previous_ext_re.group(5)
        self._need_redo_archive(force, old_version, old_extension)

        return specfile

    def _check_archive_specfile(self, specfile):
        # Ensure the archive is present in specfile
        for source in specfile.sources:
            if os.path.basename(replace_macros(source, specfile)).startswith(replace_macros(self._archive_pattern, specfile)):
                self._archive_extension = os.path.basename(replace_macros(source, specfile)).replace(
                    replace_macros(self._archive_pattern, specfile), "")
                break
        if not self._archive_extension or not self._archive_extension.startswith("."):
            raise NoArchiveNameVersionError("No archive find in specfile with the pattern {}".format(
                self._archive_pattern
            ))

    def adjust_tagv(self, tagv):
        tag_re = re.fullmatch(self.REGEX_VERSION_SPECFILE, tagv)
        if tag_re:
            return tagv

        # version is 1 digit: X
        tag_re = re.fullmatch(r"\d+", tagv)
        if tag_re:
            return "{}.0.0".format(tagv)

        # version is 2 digit: X
        tag_re = re.fullmatch(r"\d+\.\d+", tagv)
        if tag_re:
            return "{}.0".format(tagv)

        raise GitSourceError("tagv {} is not matching with regex: {}".format(
            tagv, self.REGEX_VERSION_SPECFILE))

    def _set_extended_version_tag(
            self, version, nb_commit, gcommit, detailed=False):
        commit_date = "{}{:02d}{:02d}".format(
            self.repo_gitsource.head.commit.committed_datetime.year,
            self.repo_gitsource.head.commit.committed_datetime.month,
            self.repo_gitsource.head.commit.committed_datetime.day)
        if detailed:
            commit_date = "{}+{:02d}{:02d}{:02d}".format(
                commit_date,
                self.repo_gitsource.head.commit.committed_datetime.hour,
                self.repo_gitsource.head.commit.committed_datetime.minute,
                self.repo_gitsource.head.commit.committed_datetime.second)

        self._version_tag = "{}+{}+{}+g{}".format(
            version, commit_date, nb_commit, gcommit)

    def _get_true_tag_from_tag_to_exclude(self, tag_to_exclude, nb_commit):
        # retrieve the tag object to ensure tag_to_exclude is inside tag object
        # it is a git issue: tag name mismatch tag path!
        if not nb_commit:
            nb_commit = 0
        i = 0

        # find the commits
        commit_parent = self.repo_gitsource.head.commit
        while i < nb_commit:
            if not commit_parent.parents:
                return None
            commit_parent = commit_parent.parents[0]
            i += 1

        true_tags = self.repo_gitsource.git.tag(points_at=commit_parent.hexsha)
        if not true_tags:
            return None
        for true_tag in true_tags.split('\n'):
            false_tag = self.repo_gitsource.git.tag(
                format="%(tag)", l=true_tag)
            if false_tag == tag_to_exclude:
                return true_tag

        return None

    def _set_version_tag_from_describe(self):
        describe = None
        if not hasattr(self, "_args_describe_exclude"):
            self._args_describe_exclude = ["--tags", "--first-parent"]

        try:
            describe = self.repo_gitsource.git.describe(
                self._args_describe_exclude)
        except git.exc.GitCommandError:
            # no decription found
            return False

        # get info from describe git matching our own criteria
        (self._version_tag, nb_commit, gcommit) = self._regex_describe(
            describe)
        self.log.debug("[describe]: %s, _version_tag:%s, nb_commit:%s, gcommit:%s",
                       describe, self._version_tag, nb_commit, gcommit)

        # if no info found try again
        if not self._version_tag:
            REGEX_GET_TAG_FROM_DESCRIBE = r"-\d+-g\w+$"
            tag_to_exclude = re.sub(REGEX_GET_TAG_FROM_DESCRIBE, "", describe)
            # check if tag exists
            tmptag = self.repo_gitsource.git.tag(l=tag_to_exclude)
            # if tag does not exists: could be an issue in git object
            # so check in the tag object
            if not tmptag:
                self.log.debug(
                    "[tag_to_exclude]: %s is not a tag", tag_to_exclude)
                tag_to_exclude = self._get_true_tag_from_tag_to_exclude(
                    tag_to_exclude, nb_commit)
                self.log.debug(
                    "[tag_to_exclude]: found true tag matching: %s", tag_to_exclude)
                if not tag_to_exclude:
                    return False

            self._args_describe_exclude += ["--exclude", tag_to_exclude]
            return self._set_version_tag_from_describe()

        # if info found but commit is not tagged
        if nb_commit and gcommit:
            self._set_extended_version_tag(
                self._version_tag, nb_commit, gcommit)
        self.log.info("Set version tag to %s", self._version_tag)
        return True

    def _looking_for_tag_with_git_describe(self):
        # looking for last commit tag
        self._version_tag = None
        if self._set_version_tag_from_describe() or not self._is_shallow_clone():
            return

        self.log.info("Not found tag matching regex: fetch deeper!")
        # shallow cloned: maybe tree is not enough complete to find a tag:
        # fetching
        self._git_fetch_depth_callback(
            200, self._set_version_tag_from_describe)

    def _get_change_log(self):
        today_str = datetime.date.today().strftime("%a %b %d %Y")
        commit_msg = ""
        lines_commit_msg = self.repo_gitsource.head.commit.message.split(
            os.linesep)
        # remove last empty element
        if lines_commit_msg[-1] == "":
            lines_commit_msg.pop(-1)
        for cmsg in lines_commit_msg:
            commit_msg += "- \t%s%s" % (cmsg, os.linesep)
        return "%s%s%s%s%s%s%s%s%s" % (os.linesep,
                                       "* %s IoT.bzh(iotpkg) <redpesk.list@iot.bzh> %s" % (
                                           today_str, self._version_tag),
                                       os.linesep,
                                       "- Upgrade version from source commit sha: %s" % self.repo_gitsource.head.commit.hexsha,
                                       os.linesep,
                                       "- Commit message:",
                                       os.linesep,
                                       commit_msg,
                                       os.linesep)

    def _get_new_spec_file_content(self, spec_content, no_changelog):
        # Remove #Hexsha from specfile
        new_spec_file_content = re.sub(
            r"\n#Hexsha:\s*(?:[a-f0-9]{7,40}|None)\s*(?=\n)",
            "", spec_content, flags=re.IGNORECASE)
        # Append #Hexsha before Version
        new_spec_file_content = re.sub(
            r"\nversion:(\s*)[\w\W]+?\n",
            "\n#Hexsha:\\g<1>{}\nVersion:\\g<1>{}\n".format(
                self.repo_gitsource.head.commit.hexsha,
                self._version_tag), new_spec_file_content, 1, re.IGNORECASE)
        # Apend changelog if needed
        if not no_changelog:
            changelog = self._get_change_log()
            if re.search(r"\n%changelog\s*\n", new_spec_file_content):
                new_spec_file_content = re.sub(
                    r"\n%changelog\s*\n",
                    "\n%changelog\n{}\n".format(changelog),
                    new_spec_file_content, 1, re.IGNORECASE)
            else:
                new_spec_file_content += "\n%changelog\n{}".format(changelog)
        return new_spec_file_content

    def _check_source(self, new_spec_file_content, archive_name):
        # check Source
        is_archive_in_sources = False
        new_spec = from_string(new_spec_file_content)
        for source in new_spec.sources:
            if archive_name == os.path.basename(replace_macros(source, new_spec)):
                is_archive_in_sources = True
        if not is_archive_in_sources:
            raise GitSourceError(
                "{} is not present in expanded sources ".format(archive_name))

    def _get_format(self):
        ok = False
        extensions = {elem[0]: elem[1] for elem in shutil.get_unpack_formats()}
        for _key in extensions:
            if self._archive_extension in extensions[_key]:
                ok = True
                break
        if ok:
            return _key
        else:
            raise GitSourceError(
                "Archive extension {} is not supported".format(
                    self._archive_extension))

        return (first_extension, second_extension)

    def _make_archive(self):
        # archive
        self.log.info("Making archive ...")
        format = self._get_format()
        prefix = "{}-{}".format(replace_macros(self._spec.name, self._spec), self._version_tag)
        prefix_tmp = "{}~".format(prefix)
        archive_name = "{}{}".format(prefix, self._archive_extension)
        archive_path = os.path.join(self._workdir, archive_name)
        # Check for sources submodule
        if self.repo_gitsource.submodules:
            self.log.debug("Submodules contained")
            self.repo_gitsource.git.submodule("update", "--init", "--recursive")
        with self.rename_gitsource(prefix):
            # Execute shutil.make_archive in a different process because a os.chdir is done and has side effects on other threads
            with concurrent.futures.ProcessPoolExecutor(1) as executor:
                future = executor.submit(shutil.make_archive, os.path.join(self._workdir, prefix_tmp), format, self._workdir, prefix)
                return (archive_name, archive_path, future.result(3600))

    def gitsource(self, spec_content=None, commit=None,
                  path_source=None, clean=False):

        self._tags = {}
        self._archive_extension = None
        self._set_spec(spec_content)
        self._set_path_source(path_source)
        self._set_default_branch()
        self._set_commit_checkout(commit)
        self._set_ref_type()

        self.repo_gitsource = None
        # if local repository
        if os.path.isdir(self._path_source):
            self._set_gitsource_localdir(clean)
        elif os.path.isdir(self._gitsourcedir):
            clean = self._set_gitsource(clean)

        # clean source
        if clean:
            if os.path.isdir(self._gitsourcedir):
                self.log.debug("Remove %s", self._gitsourcedir)
                shutil.rmtree(self._gitsourcedir)
                self.repo_gitsource = None

        os.makedirs(self._gitsourcedir, exist_ok=True)

        # gitsource does not exist, clone the repo into it
        if not self.repo_gitsource:
            self._shallow_clone()
        # gitsource exists, update it
        else:
            if not self.checkout_success():
                self._git_fetch_depth(1)
            else:
                self._get_remote_origin().pull(self._commit_checkout, quiet=True, prune=True, rebase=True)
            self._go_checkout()
        return self.repo_gitsource.head.commit

    def archive(self, spec_content, version=None, path_source=None,
                commit=None, clean=False, no_changelog=True,
                previous_archive_name=None, force=False):
        if spec_content and version:
            specfile = from_string(spec_content)
            version = replace_macros(version, specfile)

        self.gitsource(spec_content, path_source=path_source,
                       commit=commit, clean=clean)
        self.log.debug("Current commit is %s", self.repo_gitsource.head.commit)
        if version:
            self._version_tag = version
        elif self._git_check_ls_remote_match_tag_regex():
            self._looking_for_tag_with_git_describe()
        else:
            self.log.debug(
                "None of the tags in the repository match the regex")
            self._version_tag = None

        # no tag found: version is 0.0.0
        if not self._version_tag:
            self._set_extended_version_tag(
                "0.0.0", "0", self.repo_gitsource.head.commit.hexsha[0:7], True)

        self._set_specfile(spec_content, previous_archive_name, force)

        self.log.debug("Version is %s", self._version_tag)

        (archive_name, archive_path, archive_path_tmp) = self._make_archive()

        new_spec_file_content = self._get_new_spec_file_content(
            spec_content, no_changelog)

        self._check_source(new_spec_file_content, archive_name)

        # rename temporary archive
        if os.path.isfile(archive_path):
            os.remove(archive_path)
        os.rename(archive_path_tmp, archive_path)

        self.log.info(archive_name)
        return archive_path, new_spec_file_content, self.repo_gitsource.head.commit.hexsha

    def exclude_git(self, filename: str):
        return filename == os.path.join(self._gitsourcedir, ".git")

    @contextlib.contextmanager
    def rename_gitsource(self, prefix: str):
        with tempfile.TemporaryDirectory() as tmp_dir:
            shutil.move(os.path.join(self._gitsourcedir, ".git"), tmp_dir)
            shutil.move(self._gitsourcedir, os.path.join(self._workdir, prefix))
            try: yield
            finally:
                shutil.move(os.path.join(self._workdir, prefix), self._gitsourcedir)
                shutil.move(os.path.join(tmp_dir, ".git"), self._gitsourcedir)
