
import aiohttp
import asyncio
import json
import functools
import socketio
from typing import Dict

from .log import red_log

API_HELP = "/help"
API_VERSION = "/version"
API_HEALTHZ = "/healthz"
API_EVENT = "/events/register"


class HttpCliError(Exception):
    pass


class HttpCliReqError(Exception):
    pass

class HttpCliEventError(Exception):
    pass

class EventCliName:
    """class that handle a queue by event"""
    def __init__(self, name):
        self._name = name
        self._queues: Dict[str, asyncio.Queue] = {}
        self._callbacks = {}

    @property
    def name(self):
        return self._name

    def get_queue(self, ticket):
        queue = self._queues.get(ticket, None)
        if not queue:
            queue = asyncio.Queue(maxsize=5)
            self._queues[ticket] = queue
        return queue

    async def handler(self, data):
        ticketid = data.get("ticket", None)
        if not ticketid:
            red_log.warning("No ticket id for event=%s data=%s", self._name, data)
            return

        red_log.debug("received %s", data)
        await self.get_queue(ticketid).put(data)

    async def wait_ticket(self, ticket):
        q = self.get_queue(ticket)
        data = await q.get()
        q.task_done()
        return data


class ClientSessionFixUri(aiohttp.ClientSession):
    """in python3.6 there is no uri in object param"""
    def __init__(self, uri, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._uri = uri

    def get(self, route, *args, **kwargs):
        return super().get(self._uri + route, *args, **kwargs)

    def post(self, route, *args, **kwargs):
        return super().post(self._uri + route, *args, **kwargs)

    def put(self, route, *args, **kwargs):
        return super().put(self._uri + route, *args, **kwargs)

    def delete(self, route, *args, **kwargs):
        return super().delete(self._uri + route, *args, **kwargs)


def retry():
    """decorator that try several times request to microservice analyser if issue"""
    def decorator(method):
        async def wrapper(self, *args, **kwargs):
            cpt = 5
            while True:
                try:
                    return await method(self, *args, **kwargs)
                except HttpCliError as e:
                    red_log.warning("Issue calling request error=%s", e)
                    raise e
                except Exception as e:
                    cpt -= 1
                    red_log.warning("Issue calling request error=%s, retrying %s times", e, cpt)
                    await asyncio.sleep(1)
                    if cpt <= 0:
                        if self._connect_task.done():
                            self._connect_task = asyncio.create_task(self.connect())
                        raise HttpCliError(f"Cannot connect to httpcli error={e}")

        return functools.update_wrapper(wrapper, method)
    return decorator


class HttpEventCli:
    """class that wrap basic http request and events client behavior"""
    def __init__(self, uri, cookie_name, key_event="evtName", eventstoregister=[], logger=False) -> None:
        self._uri = uri
        self._cookie_name = cookie_name
        self._retry = 5
        self._connected = False
        self._sioconnected = False

        self._session = ClientSessionFixUri(self._uri, json_serialize=json.dumps)
        self._sio = socketio.AsyncClient(reconnection=True, logger=logger, engineio_logger=logger)

        self._key_event = key_event
        self._eventstoregister = eventstoregister
        self._events: Dict[str, EventCliName] = {}

        self._connect_task = asyncio.ensure_future(self.connect())

    @property
    def session(self) -> ClientSessionFixUri:
        return self._session

    @property
    def sio(self) -> socketio.AsyncClient:
        return self._sio

    def stop_background_tasks(self):
        """cancel background tasks, just need to wait them"""
        self._connect_task.cancel()
        return [self._connect_task]

    async def sioconnect(self, cookie):
        if not self._sioconnected:
            headers = {"Cookie": cookie.key + "=" + cookie.value}
            await self._sio.connect(self._uri, headers=headers, transports=['websocket'])
            #self._sio.ping_timeout = 60*2  # put timeout at 2 minutes
            red_log.info("[EventClient]: connected to {} with cookie {}={}".format(self._uri, cookie.key, cookie.value))
            self._sioconnected = True
            self._sio.on("connect", self._connect)
            self._sio.on("disconnect", self._disconnect)

    async def register_events(self):
        return await asyncio.gather(*[self.event_register(evt) for evt in self._eventstoregister])

    async def connect(self):
        red_log.info("Try to connect to %s", self._uri)

        while not self._connected:
            try:
                async with self._session.get(API_VERSION) as resp:
                    if resp.status != 200:
                        raise HttpCliError("Cannot get version from %s %s", self._uri, resp)
                    red_log.info("Connected to %s version=%s", self._uri, json.loads(await resp.text()))

                    await self.sioconnect(resp.cookies[self._cookie_name])

            except Exception as e:
                red_log.warning("Cannot connect to %s error=%s, retrying in %ss",
                                self._uri, e, self._retry)
            else:
                self._connected = True
                # register to all evts
                await self.register_events()
            await asyncio.sleep(self._retry)

    async def _handle_response(self, r):
        res = await r.json()
        if r.status != 200:
            raise HttpCliReqError("Issue {}: {}".format(self._uri, res))
        return res

    @retry()
    async def get(self, route):
        async with self._session.get(route) as r:
            return await self._handle_response(r)

    async def _be_sure_events_registered(self, route):
        if route == API_EVENT:
            return

        await self.register_events()

    @retry()
    async def post(self, route, data):
        await self._be_sure_events_registered(route)
        async with self._session.post(route, json=data) as r:
            return await self._handle_response(r)

    @retry()
    async def put(self, route, data):
        await self._be_sure_events_registered(route)
        async with self._session.delete(route, json=data) as r:
            return await self._handle_response(r)

    @retry()
    async def delete(self, route, data=None):
        await self._be_sure_events_registered(route)
        async with self._session.delete(route, json=data) as r:
            return await self._handle_response(r)

    async def version(self):
        """get service version"""
        return self.get(API_VERSION)

    async def event_register(self, event_name):
        json_data = {self._key_event: event_name}
        await self.post(API_EVENT, json_data)
        self.sioregister(event_name)

    async def event_unregister(self, event_name):
        return await self.delete(f"{API_EVENT}/{event_name}")

    def sioregister(self, event_name):
        """register to sio event"""
        event = self._events.get(event_name, None)
        if not event:
            event = EventCliName(event_name)
            self._events[event_name] = event

        self._sio.on(event.name, handler=event.handler)
        red_log.info("[sioregister]: register to event=%s", event.name)

    async def _connect(self):
        red_log.info("Connected!")

    async def _disconnect(self):
        if self._sioconnected:
            red_log.info("disconnected from server")
            await self._sio.disconnect()
            self._sioconnected = False
            self._connected = False
            self._connect_task = asyncio.ensure_future(self.connect())

    async def wait_ticket(self, event_name, ticket, strict=True):
        data = {}
        done = False
        while not done:
            data = await self._events[event_name].wait_ticket(ticket)
            done = data["done"]

        # handle ticket error
        if data.get("error", None):
            error_msg = f"Issue wait ticket for event={event_name} res={res} error={error}"
            red_log.warning(error_msg)
            if strict:
                raise HttpCliEventError(error_msg)

        red_log.debug("[wait_ticket]: received ticket %s", data)
        return data
